package input;

import java.util.Vector;

/**
 * Reprsentation simplifiee d'une classe.
 * Elles sont publiques
 * @author jroyer
 *
 */
public class Class {

	/**
	 * Constante utile pour l'absence d'attribut.
	 */
	public static final Vector<Attribute> NOATTRIBUTE = new Vector<Attribute>();

	/**
	 * Le nom de la classe.
	 */
	protected String name;

	/**
	 * Ses attributs.
	 */
	protected Vector<Attribute> structure;

	/**
	 * Ses methodes.
	 */
	protected Vector<Method> behaviour;

	/**
	 * Abstraite ou non.
	 */
	protected boolean concrete;

	/**
	 * Super classe ou null sinon.
	 */
	protected String inherits;

	/**
	 * Constructeur d'une classe.
	 * @param n nom de la classe
	 * @param attr les attributs
	 * @param beh les methodes
	 * @param a concrete ou abstraite
	 * @param s nom super classe
	 */
	public Class(String n, Vector<Attribute> attr, Vector<Method> beh, boolean a, String s) {
		this.name = n;
		this.structure = attr;
		this.behaviour = beh;
		this.concrete = a;
		this.inherits = s;
	}

	/**
	 * Representation d'une classe.
	 * @return une chaine
	 */
	public String toString() {
		return "Classe : " + this.getName() + "\n"
		+ "Concrete : " + this.isConcrete() + "\n"
		+ "Herite : " + this.getInherits() + "\n"
		+ "Attributs : " + this.getStructure() + "\n"
		+ "Comportements : " + this.getBehaviour() + "\n";
	}

	//------------- generation du code --------
	public String generate()  {
		String result = Keywords.PUBLIC;
		if (! this.isConcrete()) {
			result += Keywords.ABSTRACT;
		}
		result += Keywords.CLASS + this.getName() + Keywords.BLANC;
		if (this.getInherits() != null) {
			result += Keywords.EXTENDS + this.getInherits();
		}
		result += Keywords.ACCIN + Keywords.NEWLINE + Keywords.NEWLINE;
		// les attributs
		for(Attribute at : this.getStructure()) {
			result += at.generate();
		}
		// le constructeur
		if (this.isConcrete()) {
			result += Keywords.NEWLINE;
			result += this.generateConstructor();
		}
		// NEW
		// copy constructor
		if (this.isConcrete() && !this.structure.isEmpty()) {
			result += Keywords.NEWLINE;
			result += this.generateCopyConstructor();
		}
		// les methodes
		for(Method met : this.getBehaviour()) {
			result += Keywords.NEWLINE + Keywords.NEWLINE;
			result += met.generate();
		}
		// les accesseurs
		for(Attribute at : this.getStructure()) {
			result += Keywords.NEWLINE + Keywords.NEWLINE;
			result += at.generateAccessors();
		}
		// equals
		result += Keywords.NEWLINE + Keywords.NEWLINE;
		result += this.generateEquals();
		result += Keywords.NEWLINE + Keywords.ACCOUT + Keywords.NEWLINE;
		return result;
	}

	/**
	*
	 * @return Genere le code d'un constructeur complet.
	 */
	public String generateConstructor() {
		String towrite = Keywords.TAB + Keywords.COMMENT
						+ "Le constructeur complet." + Keywords.NEWLINE;
		towrite += Keywords.TAB + Keywords.PUBLIC + this.getName() + Keywords.PARIN;
		// les parametres
		int length = this.getStructure().size();
		if (length > 0) {
			towrite += this.getStructure().get(0).generateParameter();
			for (int i = 1; i < length; i++) {
				towrite += Keywords.COLON + this.getStructure().get(i).generateParameter();
			}
		}
		towrite += Keywords.PAROUT + Keywords.ACCIN + Keywords.NEWLINE;
		// body
		for (Attribute attr : this.getStructure()) {
			towrite += Keywords.TAB + Keywords.TAB
					+ Keywords.THIS + Keywords.DOT + attr.getName()
					+ Keywords.SET + attr.getName()
					+ Keywords.SEMICOLON + Keywords.NEWLINE;
		}
		return towrite + Keywords.TAB + Keywords.ACCOUT + Keywords.NEWLINE;
	}

	/**
	 * @return Genere le code d'un constructeur de recopie superficielle.
	 * Construit un constructeur de recopie sans appel au constructeur complet.
	 */
/*	public String generateCopyConstructor() {
		String towrite = Keywords.TAB + Keywords.COMMENT
						+ "Le constucteur de recopie superficielle." + Keywords.NEWLINE;
		towrite += Keywords.TAB + Keywords.PUBLIC + this.getName() + Keywords.PARIN;
		// les parametres : ClassName other
		towrite += this.name + Keywords.BLANC + Keywords.OTHER;
		towrite += Keywords.PAROUT + Keywords.ACCIN + Keywords.NEWLINE;
		// body
		for (Attribute attr : this.getStructure()) {
			towrite += Keywords.TAB + Keywords.TAB
					+ Keywords.THIS + Keywords.DOT + attr.getName()
					+ Keywords.SET + Keywords.BLANC + Keywords.OTHER + Keywords.DOT + attr.getName()
					+ Keywords.SEMICOLON + Keywords.NEWLINE;
		}
		return towrite + Keywords.TAB + Keywords.ACCOUT + Keywords.NEWLINE;
	}*/

	/**
	 * @return Genere le code d'un constructeur de recopie superficielle.
	 * mais avec appel du constructeur comple
	 */
	public String generateCopyConstructor() {
		String towrite = Keywords.TAB + Keywords.COMMENT
						+ "Le constructeur de recopie superficielle avec appel a this()." + Keywords.NEWLINE;
		towrite += Keywords.TAB + Keywords.PUBLIC + this.getName() + Keywords.PARIN;
		// les parametres : ClassName other
		towrite += this.name + Keywords.BLANC + Keywords.OTHER;
		towrite += Keywords.PAROUT + Keywords.ACCIN + Keywords.NEWLINE;
		// body = this( other.att *)
		towrite += Keywords.TAB +  Keywords.TAB + Keywords.THIS + Keywords.PARIN ;
		int length = this.getStructure().size();
		if (length > 0) {
			towrite += Keywords.OTHER + Keywords.DOT + this.getStructure().get(0).getName();
			for (int i = 1; i < length; i++) {
				towrite += Keywords.COLON + Keywords.OTHER + Keywords.DOT + this.getStructure().get(i).getName();
			}
		}
		towrite += Keywords.PAROUT + Keywords.SEMICOLON  + Keywords.NEWLINE;
		return towrite +  Keywords.TAB + Keywords.ACCOUT + Keywords.NEWLINE;
	} // end generateCopyConstructor

	/**
	 * Generate the equals method.
	 * @return
	 */
	public String generateEquals() {
		// le commentaire
		String towrite = Keywords.TAB + Keywords.COMMENT
				+ "Deep equality test." + Keywords.NEWLINE;
		// la visibilite
		towrite += Keywords.TAB + Keywords.PUBLIC;
		towrite += Keywords.BOOLEAN + Keywords.BLANC + "equals " + Keywords.PARIN;
		// les parametres
		towrite += "Object " + Keywords.OTHER;
		towrite += Keywords.PAROUT + Keywords.ACCIN + Keywords.NEWLINE;
		towrite += Keywords.TAB +  Keywords.TAB ;
		// body
		towrite +=  Keywords.RETURN + "(other != null) " + Keywords.AND
			+ Keywords.PARIN + "other instanceof " + this.name  + Keywords.PAROUT;
		for (Attribute attr : this.getStructure()) {
			towrite += Keywords.AND + Keywords.NEWLINE +  Keywords.TAB + Keywords.TAB;
			// must distinguish primitive types from others
			// == or equals(...)
			if (Primitive.isPrimitive(attr.getType())) {
				towrite += Keywords.PARIN +  Keywords.THIS + Keywords.DOT + attr.getName() +
							Keywords.EQUAL +
							// cast needed
							 Keywords.PARIN + Keywords.PARIN + this.name  + Keywords.PAROUT
							     +  Keywords.OTHER + Keywords.PAROUT + Keywords.DOT +  attr.getName() + Keywords.PAROUT;
			} else {
				towrite += Keywords.THIS + Keywords.DOT + attr.getName() + Keywords.DOT + "equals" + Keywords.PARIN
				// cast needed
				+ Keywords.PARIN + Keywords.PARIN + this.name  + Keywords.PAROUT + Keywords.OTHER + Keywords.PAROUT
				+ Keywords.DOT +  attr.getName() + Keywords.PAROUT;
			}
		}
		towrite += Keywords.SEMICOLON  + Keywords.NEWLINE;
		towrite += Keywords.TAB + Keywords.ACCOUT;
		return towrite;
	}

	/**
	 *
	 * @return le vecteur des attributs.
	 */
	public Vector<Attribute> getStructure() {
		return structure;
	}

	/**
	 * Affecte le vecteur des attributs.
	 * @param structure
	 */
	public void setStructure(final Vector<Attribute> structure) {
		this.structure = structure;
	}

	/**
	 *
	 * @return le vecteur des methodes.
	 */
	public Vector<Method> getBehaviour() {
		return behaviour;
	}

	/**
	 * Affecte le vecteur des methodes.
	 * @param behaviour
	 */
	public void setBehaviour(Vector<Method> behaviour) {
		this.behaviour = behaviour;
	}

	/**
	 *
	 * @return false si la classe est abstraite.
	 */
	public boolean isConcrete() {
		return concrete;
	}

	/**
	 *
	 * @return le nom de la classe.
	 */
	public String getName() {
		return name;
	}

	/**
	 * Affecte le nom de la classe.
	 * @param name
	 */
	public void setName(String name) {
		this.name = name;
	}

	/**
	 *
	 * @return null ou la super classe.
	 */
	public String getInherits() {
		return inherits;
	}

	/**
	 * Affecte la super classe.
	 * @param inherits
	 */
	public void setInherits(String inherits) {
		this.inherits = inherits;
	}

	/**
	 * Affecte le statut concret.
	 * @param concrete
	 */
	public void setConcrete(boolean concrete) {
		this.concrete = concrete;
	}

}
